package com.malk.poc.service.impl;

import cn.hutool.core.util.ObjectUtil;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.google.common.collect.Lists;
import com.malk.poc.server.AWServer;
import com.malk.poc.server.YDSearch;
import com.malk.poc.server.YdIpdForm;
import com.malk.poc.server.YdProject;
import com.malk.poc.service.AWClint;
import com.malk.poc.service.AwDingService;
import com.malk.server.aliwork.YDConf;
import com.malk.server.aliwork.YDParam;
import com.malk.server.common.McException;
import com.malk.server.dingtalk.DDConf;
import com.malk.server.dingtalk.DDR_New;
import com.malk.server.teambition.TBConf;
import com.malk.service.aliwork.YDClient;
import com.malk.service.aliwork.YDService;
import com.malk.service.dingtalk.*;
import com.malk.service.teambition.TBClient;
import com.malk.utils.*;
import lombok.Synchronized;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.*;
import java.util.stream.Collectors;

@Service
@Slf4j
public class AWImplClient implements AWClint {

    @Autowired
    private YDClient ydClient;
    @Autowired
    private YDService ydService;
    @Autowired
    private YdProject ydProject;
    @Autowired
    private YdIpdForm ydIpdForm;
    @Autowired
    private TBClient tbClient;
    @Autowired
    private TBConf tbConf;
    @Autowired
    private DDConf ddConf;
    @Autowired
    private DDClient ddClient;
    @Autowired
    private DDClient_Contacts ddClient_contacts;
    @Autowired
    private DDClient_Storage ddClient_storage;
    @Autowired
    private DDClient_Notice ddClientNotice;

    /**
     * 更新任务自定义审批字段值
     *
     * @apiNote https://open.teambition.com/docs/documents/664dad5aba1c5ca0e1576f8f#4%230%23%E5%AE%A1%E6%89%B9%E5%8A%9F%E8%83%BD%E9%85%8D%E7%BD%AE
     * [ ppExt ]
     * 1. 更新需要为 metaString, 若是 meta 自定义字段不生效. 目前发现仅自定义字段有此问题
     * 2. tb 内弹出框打开可用 handler, 但宜搭链接会被拦击, 直接赋值url新标签页打开
     * 3. 标题字段直接赋值, 不能添加自定义值, 如标记颜色, 否则点击连接诶后任务上该字段会消失: meta.put("title", UtilMap.map("color, value", "BLUE", "xxx"))
     */
    private void _updateApproveField(String taskId, List<Map> approves, String url, String title, String subTitle, String fieldName) {
        Map colors = UtilMap.map("待提交, 已提交, 审批中, 审批通过, 审批拒绝, 已终止", "ORANGE, BLUE, BLUE, GREEN, RED, RED");
        Map meta = UtilMap.map("instanceId, url", taskId, url);
        meta.put("icon", UtilMap.map("value, color, bgColor", "Icon_Approve, WHITE, ORANGE"));
        meta.put("subTitle", UtilMap.map("value, color", subTitle, colors.get(subTitle)));
        approves.add(UtilMap.map("title, metaString", title, JSON.toJSONString(meta)));
        log.info("更新任务自定义审批字段值, {}, {}", taskId, JSON.toJSONString(approves));
        tbClient.updateTaskCustomField(taskId, tbConf.getOperatorId(), UtilMap.map("customfieldName, value", fieldName, approves));
    }

    // 获取任务详情 fixme 提取到 tbservice
    private Map<String, String> _getTaskFieldMap(Map task, List<Map> projectCustomField, String... fieldNames) {
        return __getTaskMap(task, projectCustomField, fieldNames);
    }

    // 组装任务数据
    private Map __getTaskMap(Map task, List<Map> projectCustomField, String... fieldNames) {
        Map result = UtilMap.map("task", task);
        List<Map> taskCustomField = (List<Map>) task.get("customfields");
        for (String filed : fieldNames) {
            Optional optional = projectCustomField.stream().filter(item -> filed.equals(item.get("name"))).findAny();
            if (optional.isPresent()) {
                Map map = (Map) optional.get();
                result.put(filed, TBConf.getTaskFieldValue_First(taskCustomField, String.valueOf(map.get("id"))));
                result.put(filed + "_id", map.get("id")); // ppExt: 返回字段ID
            }
        }
        return result;
    }

    // 获取任务详情 fixme 提取到 tbservice
    private Map _getTaskFieldMap(String taskId, String... fieldNames) {
        // 查询任务详情
        Map rTask = tbClient.queryTaskDetail(taskId, null, null).get(0);
        // 查询项目任务ID
        List<Map> customField = tbClient.queryProjectCustomField(String.valueOf(rTask.get("projectId")), null);
        return __getTaskMap(rTask, customField, fieldNames);
    }

    // 匹配工作流名称, 获取ID
    @Deprecated
    String _getWorkFlowStatus(String projectId, String workFlowStatusName) {

        List<Map> customFlowStatus = tbClient.queryProjectCustomFlowStatus(projectId, UtilMap.map("q", workFlowStatusName));
        // 7.10 避免模糊匹配
        customFlowStatus = customFlowStatus.stream().filter(item -> workFlowStatusName.equals(UtilMap.getString(item, "name"))).collect(Collectors.toList());
        McException.assertAccessException(customFlowStatus.isEmpty(), "工作流名称未匹配");
        return String.valueOf(customFlowStatus.get(0).get("id"));
    }

    String _getWorkFlow(String projectId, String workFlowName) {

        List<Map> customFlowStatus = tbClient.queryProjectCustomFlow(projectId, UtilMap.map("q", workFlowName));
        // 7.10 避免模糊匹配
        customFlowStatus = customFlowStatus.stream().filter(item -> workFlowName.equals(UtilMap.getString(item, "name"))).collect(Collectors.toList());
        McException.assertAccessException(customFlowStatus.isEmpty(), "工作流名称未匹配");
        return String.valueOf(customFlowStatus.get(0).get("id"));
    }

    String _getTaskType(String projectId, String taskTypeName) {

        List<Map> customFlowStatus = tbClient.queryProjectTaskType(projectId, UtilMap.map("q", taskTypeName));
        // 7.10 避免模糊匹配
        customFlowStatus = customFlowStatus.stream().filter(item -> taskTypeName.equals(UtilMap.getString(item, "name"))).collect(Collectors.toList());
        McException.assertAccessException(customFlowStatus.isEmpty(), "工作流名称未匹配");
        return String.valueOf(customFlowStatus.get(0).get("id"));
    }

    // 前置过滤 fixme 提取方法, 优化为单个查询, 处理保密模糊匹配
    @Deprecated
    List<String> _getWorkFlowStatusList(String projectId, String... workFlowStatusNames) {
        List<String> workFlowStatusList = new ArrayList<>();
        for (String name : workFlowStatusNames) {
            List<Map> customFlowStatus = tbClient.queryProjectCustomFlowStatus(projectId, UtilMap.map("q", name));
            // 7.10 避免模糊匹配
            customFlowStatus = customFlowStatus.stream().filter(item -> name.equals(UtilMap.getString(item, "name"))).collect(Collectors.toList());
            workFlowStatusList.addAll(customFlowStatus.stream().map(item -> UtilMap.getString(item, "id")).collect(Collectors.toList()));
        }
        return workFlowStatusList;
    }

    // 后置过滤 fixme 提取方法
    List<String> _getWorkFlowStatusList(String projectId, List<String> workFlowStatusNames) {
        //List<String> names = Arrays.stream(workFlowStatusNames).collect(Collectors.toList());
        List<Map> customFlowStatus = tbClient.queryProjectCustomFlowStatus(projectId, null);
        List<Map> workFlowStatusList = customFlowStatus.stream().filter(item -> workFlowStatusNames.contains(item.get("name"))).collect(Collectors.toList());
        return workFlowStatusList.stream().map(item -> UtilMap.getString(item, "id")).collect(Collectors.toList());
    }

    // 通过id获取工作流名称
    private String _getWorkFlowStatusName(String projectId, String tfsId) {
        List<Map> customFlowStatus = tbClient.queryProjectCustomFlowStatus(projectId, UtilMap.map("tfsIds", tfsId));
        return UtilMap.getString(customFlowStatus.get(0), "name");
    }

    /**
     * 获取知识库附件传递到宜搭
     * -
     * 1. 宜搭附件传递 downloadUrl 和 name 即可实现在线预览
     * 2. 知识库绑定没有文件后缀, 宜搭识别目前仅能点击下载跳转预览, 添加统一docx后缀, 文档会自行区分
     * -
     * ppExt: 获取任务字段需要传递字段名称以获取ID, 另外添加 .doc 只是重定向若是文件夹也不影响访问
     */
    List<Map> _getDocs(Map taskData, String fieldName) {
        Map rTask = UtilMap.getMap(taskData, "task");
        List<Map> customfields = UtilMap.getList(rTask, "customfields");
        String deliverable = String.valueOf(taskData.get(fieldName + "_id")); // 交付物任务字段ID
        List<Map> attachments = TBConf.getTaskFieldValue(customfields, deliverable).stream().map(item -> {
            // 7.8 兼容知识库版本: 无编辑权限, 复制失败 任务更新
            if (UtilMap.isBlankString(item, "metaString")) {
                return item;
            }
            Map link = (Map) JSON.parse(String.valueOf(item.get("metaString")));
            Map row = UtilMap.map("downloadUrl, name", link.get("url"), link.get("title") + ".docx");
            row.putAll(link); // 保留原数据作为知识库版本管理相关参数
            return row;
        }).collect(Collectors.toList());
        return attachments;
    }


    /// 获取自定义字段数据列表
    List<Map> _getCustomFieldList(Map taskData, String fieldName) {
        Map rTask = UtilMap.getMap(taskData, "task");
        List<Map> customfields = UtilMap.getList(rTask, "customfields");
        String deliverable = String.valueOf(taskData.get(fieldName + "_id")); // 自定义组件ID
        return TBConf.getTaskFieldValue(customfields, deliverable);
    }

    public Map syncProject(String projectCode,String tbProjectId) {
        // 新增场景下, 提供实例ID回写TB项目编号
        Map formData = null;

        YDParam ydParam = YDParam.builder()
                .appType(ydProject.getAppType())
                .systemToken(ydProject.getSystemToken())
                .formUuid(ydProject.getProjectFid())
                .build();
        if (StringUtils.isNotBlank(projectCode)) {
            ydParam.setSearchFieldJson(JSON.toJSONString(UtilMap.map("textField_ljcbhzhi", projectCode))); // 项目名称
        }
        // prd 客户系统主数据, 是全量定时从mdm更新, 因此不能增量
        List<Map> mapBaseList = ydService.queryFormData_all(ydParam);
        if (StringUtils.isNotBlank(projectCode)) {
            mapBaseList = mapBaseList.stream().filter(item -> projectCode.equals(item.get("textField_ljcbhzhi"))).collect(Collectors.toList());
        }
        log.info("项目主数据, {}", mapBaseList.size());
        McException.assertAccessException(mapBaseList.isEmpty(), "未匹配项目主数据");
        // 查询项目角色
        YDParam ydParam1 = YDParam.builder()
                .formUuid(ydIpdForm.getRole())
                .build();
        List<Map> mapRole = (List<Map>) ydClient.queryData(ydParam1, YDConf.FORM_QUERY.retrieve_search_form).getData();

        // 匹配数据组件ID
        for (Map baseFormData : mapBaseList) {
            // 项目主数据
            formData = new HashMap();
            formData.put("textField_m9sfk19j", baseFormData.get("textField_ljcbhzhi"));      // 项目名称
            formData.put("textField_lrj7vnxb", baseFormData.get("textField_m9us05lk"));      // 项目号
//            formData.put("textField_lrj7vnxc", baseFormData.get("textField_le6o88w1"));      // 项目简称
            formData.put("textareaField_lrj7vnxl", baseFormData.get("textareaField_m72nusms"));  // 项目描述
            formData.put("textField_lrj7vnxf", baseFormData.get("formInstanceId"));          // 项目实例ID
            formData.put("textField_m9z110mk", baseFormData.get("textField_m9z3caby"));      // (项目分组)
            formData.put("textField_ltsdsti7", baseFormData.get("selectField_m71o2kk8"));      // 项目状态
//            formData.put("textField_ltsdsti8", baseFormData.get("textField_lo2c1f0r"));      // mPDT
//            formData.put("textField_ltsdsti9", baseFormData.get("textField_lki3egzn"));      // Charter子类
//            formData.put("textField_ltwcq7s5", baseFormData.get("textField_lki3egzm"));      // Charter
            formData.put("textField_ltwcq7s6", baseFormData.get("selectField_m7ivxz51"));      // 项目类型 (TB项目模板)
//            formData.put("textField_lwj1r7n6", baseFormData.get("selectField_m3qp1kc8"));      // 项目重要性分级
//            formData.put("textField_lwj1r7n7", baseFormData.get("selectField_m3qp1kc8"));      // 项目等级
            List<Map> details = new ArrayList<>();
            List<Map> roleMap = mapRole;
            Map projectMembers = baseFormData;
            // 项目角色详情 [prd 5.14 C类项目成员同步PDT成员主数据]
//            if ("Charter开发项目".equals(formData.get("textField_ltwcq7s6"))) {
////                roleMap = rolePDT;
//                String charter = String.valueOf(formData.get("textField_m9z110mk"));
//                YDParam ydParam2 = YDParam.builder()
//                        .appType("APP_LH2TJTIT3EWHCG084TGL")
//                        .systemToken("QYC66KA1F4ZG5RQV8HR9WBHYW03S3CQAJKDQLZ1")
//                        .formUuid("FORM-90E23DE1BBEB43759ED9517F98BA4938P61C")
//                        .searchFieldJson(JSON.toJSONString(UtilMap.map("textField_lr0dvhqh", charter)))
//                        .build();
//                List<Map> listPDT = (List<Map>) ydClient.queryData(ydParam2, YDConf.FORM_QUERY.retrieve_search_form).getData();
//                listPDT = listPDT.stream().filter(item -> charter.equals(((Map) item.get("formData")).get("textField_lr0dvhqh"))).collect(Collectors.toList());
//                if (listPDT.size() == 0) {
//                    continue;
//                }
//                projectMembers = (Map) listPDT.get(0).get("formData");
//            }
            for (Map roleData : roleMap) {
                Map roleFormData = (Map) roleData.get("formData");
                // 创建详情的角色名称和成员组件Map
                Object roleMember = projectMembers.get(roleFormData.get("textField_lr7bgi76")+"_id");
                String roleName = UtilMap.getString(roleFormData, "textField_lqxu439g");
                if (roleMember != null && roleName != null) {
                    // prd 提交代表标识, 用于同步项目角色
                    details.add(UtilMap.map("employeeField_lqxtykch, selectField_lqxu6bgf, radioField_lu8li5ie", roleMember, roleName, roleFormData.get("radioField_lu8li5ie")));
                    // prd 添加项目经理, 用于数据权限控制
                    if (AWServer.PROJECT_PM_NAME.equals(roleName)) {
                        formData.put("employeeField_ltzn872j", roleMember);
                    }
                }
            }
            formData.put("tableField_lqxtykcf", details);
            if(tbProjectId!=null&&!tbProjectId.isEmpty()){
                formData.put("textField_lqxtykce", tbProjectId);
                formData.put("textField_m9mjecgo", tbProjectId);
                formData.put("radioField_m9mjecgj","是");
            }else{
                formData.put("radioField_m9mjecgj","否");
            }
            // 通过项目号获取项目是否已存在
            List<Map> projectMaps = ydService.queryDataList_FormData(ydIpdForm.getProject(), UtilMap.map("textField_lrj7vnxb", baseFormData.get("textField_m9us05lk")));
            projectMaps = projectMaps.stream().filter(sub -> baseFormData.get("textField_m9us05lk").equals(sub.get("textField_lrj7vnxb"))).collect(Collectors.toList());
            YDParam ydParam2 = YDParam.builder()
                    .formUuid(ydIpdForm.getProject())
                    .formDataJson(JSON.toJSONString(formData))
                    .updateFormDataJson(JSON.toJSONString(formData))
                    .useLatestVersion(true)
                    .build();
            if (projectMaps.size() > 0) {
                /// 7.18 项目状态不一致执行概览字段更新, 避免tb重复更新造成操作记录
                if (!formData.get("textField_ltsdsti7").equals(projectMaps.get(0).get("textField_ltsdsti7"))) {
                    formData.put("src_state", formData.get("textField_ltsdsti7"));
                }
                /// 8.21 项目信息\项目重要性分级, 也添加差异同步, 避免新增字段历史项目不同步问题
                if (!formData.get("textareaField_lrj7vnxl").equals(projectMaps.get(0).get("textareaField_lrj7vnxl"))) {
                    formData.put("src_desc", formData.get("textareaField_lrj7vnxl"));
                }
//                if (!formData.get("textField_lwj1r7n6").equals(projectMaps.get(0).get("textField_lwj1r7n6"))) {
//                    formData.put("src_level", formData.get("textField_lwj1r7n6"));
//                }

                ydParam2.setFormInstanceId(String.valueOf(projectMaps.get(0).get("instanceId")));
                ydClient.operateData(ydParam2, YDConf.FORM_OPERATION.update); // 主表为增量更新
                // 保留tb项目号, 避免冲重复通过模板参加项目 [供新增用]
                formData.put("textField_lqxtykce", UtilMap.getString(projectMaps.get(0), "textField_lqxtykce"));
                formData.put("formInstanceId", projectMaps.get(0).get("instanceId"));
            } else {
                String formInstId = (String) ydClient.operateData(ydParam2, YDConf.FORM_OPERATION.create);
                formData.put("formInstanceId", formInstId);
            }
            // 全量兼容: tb项目成员更新 [兼容存量创建项目, 被删除后执行异常]
            String projectId = UtilMap.getString(formData, "textField_lqxtykce");
            if (StringUtils.isNotBlank(projectId) && StringUtils.isBlank(projectCode)) {
                try {
                    _syncProjectRole(formData, projectId, false);
                    _assembleBreakPermission(formData);
                } catch (Exception e) {
                    // 记录错误信息
                    e.printStackTrace();
                }
            }
            log.info("同步项目主数据, code = {}, id = {}", projectCode, projectCode);
        }
        return formData;
    }

    /**
     * 通过模板创建项目 [ppExt: 过滤宜搭项目档案\TB项目, 避免重复创建, 若判定为更新下执行项目成员更新]
     */
    @Override
    @Synchronized
    public void createProject(String projectCode, String templateId,String tbProjectId,String user) {

        log.info("通过模板创建项目, {}, {}", projectCode, templateId);
        // 虽然先创建TB, 可以一次性同步项目数据, 但避免执行异常, 因此成功创建TB项目再回写
        Map formData = syncProject(projectCode,tbProjectId);
        String projectId = UtilMap.getString(formData, "textField_lqxtykce");
        log.info("项目主数据, {}, {}", projectId, formData);
        boolean isCreate = StringUtils.isBlank(projectId);
        if (isCreate) {
            // prd 多模板适配: [templateId 为空, 触发项目类型匹配]
            if (StringUtils.isBlank(templateId)) {
                List<Map> dataList = ydService.queryDataList_FormData(ydIpdForm.getTemplate(), null);
                Optional optional = dataList.stream().filter(item -> item.get("selectField_ltwcsr8w").equals(formData.get("textField_ltwcq7s6"))).findAny();
                McException.assertAccessException(!optional.isPresent(), projectCode + ": TB项目模板未维护!");
                templateId = UtilMap.getString((Map) optional.get(), "textField_ltwcybz1");
                McException.assertAccessException(templateId.length() <= 1, projectCode + ": TB项目模板未匹配!");
            }
            // 通过模板创建项目, 创建项目并更新项目TB项目ID
            Map result = tbClient.projectCreateWithTemplate(projectCode, templateId, tbConf.getOperatorId());
            projectId = UtilMap.getString(result, "id");
            ydClient.operateData(YDParam.builder()
                    .formInstanceId(UtilMap.getString(formData, "formInstanceId"))
                    .updateFormDataJson(JSON.toJSONString(UtilMap.map("textField_lqxtykce", projectId)))
                    .build(), YDConf.FORM_OPERATION.update);
            log.info("TB项目号更新, {}, {}", projectCode, UtilMap.getString(formData, "formInstanceId"), projectId);
        }
        _syncProjectRole(formData, projectId, isCreate);

        // 角色自动下发
//        updateProjectRole(projectId, Arrays.asList("全部"), "","");
        if(user!=null&&!user.isEmpty()){
            String url = "https://www.teambition.com/project/" + projectId;
            Map linkMap=new HashMap();
            linkMap.put("messageUrl",url);
            linkMap.put("picUrl","@lALOACZwe2Rk");
            linkMap.put("title","项目创建完成提醒");
            linkMap.put("text","项目"+projectCode+"创建完成");
            Map msgMap=new HashMap();
            msgMap.put("msgtype","link");
            msgMap.put("link",linkMap);
            ddClientNotice.sendNotification(ddClient.getAccessToken(),Arrays.asList(user.split(",")),null,false,msgMap);
        }
    }

    // 同步TB项目成员/项目分组/项目概览
    void _syncProjectRole(Map formData, String projectId, boolean isCreate) {

        // prd TB项目成员, 项目经理设置为管理者
        List<String> pmUserId = new ArrayList<>(); // 项目经理
        List<String> zjUserId = new ArrayList<>(); // 项目总监
        List<String> roleIds = new ArrayList<>();
        List<String> representativeIds = new ArrayList<>();

//        String representative = _getProjectRoleId(projectId, "项目各代表");
        // prd 6.26 L类\C类项目使用 项目各代表_L, TB配置拥有邀请成员权限
//        if (Arrays.asList("生命周期维护项目", "Charter开发项目").contains(formData.get("textField_ltwcq7s6"))) {
//            representative = _getProjectRoleId(projectId, "项目各代表_L");
//        }
        String pmRoleId = _getProjectRoleId(projectId, AWServer.PROJECT_PM_ROLE);
        String staffRoleId = _getProjectRoleId(projectId, "项目成员");

        List<Map> details = (List<Map>) formData.get("tableField_lqxtykcf");
        for (Map row : details) {
            // 同步项目主数据, 会重新匹配项目成员, 因此无需 employeeField_xxx_id 取值
            List<String> userIds = (List<String>) UtilMap.getList(row, "employeeField_lqxtykch");
            if (userIds.isEmpty()) {
                continue;
            }
            String roleName = UtilMap.getString(row, "selectField_lqxu6bgf");
            // prd 7.3 C类项目管理员权限, 分配为产品代表
//            String tbRole = "Charter开发项目".equals(formData.get("textField_ltwcq7s6")) ? "产品代表" : AWServer.PROJECT_PM_NAME;
            if (AWServer.PROJECT_PM_NAME.equals(roleName)) {
                pmUserId.addAll(userIds);
            }
            if (AWServer.PROJECT_ZJ_NAME.equals(roleName)) {
                zjUserId.addAll(userIds);
            }
//            // prd 项目各代表, 可独立N/A任务, 添加角色控制权限
//            if ("是".equals(row.get("radioField_lu8li5ie"))) {
//                representativeIds.addAll(userIds);
//            }
            roleIds.addAll(userIds);
        }
        // 钉钉人员ID转TB人员ID
        roleIds = _convertUserId(roleIds, false);
        tbClient.createProjectMember(projectId, roleIds, tbConf.getOperatorId());
//        if (!representativeIds.isEmpty()) {
//            tbClient.updateProjectMember(_convertUserId(representativeIds, false), Arrays.asList(staffRoleId, representative), projectId);
//        }
        // prd 不做删除, 离职自动退出, 避免临时交接不可见的问题. 但需要移除项目经理权限
        List<Map> roleList = tbClient.queryProjectMember(projectId, UtilMap.map("projectRoleId", pmRoleId));
        for (Map role : roleList) {
            List<String> tRoleIds = UtilMap.getList(role, "roleIds");
            tRoleIds.remove(pmRoleId);
            if (tRoleIds.isEmpty()) {
                tRoleIds.add(staffRoleId);
            }
            if(role.containsKey("userId")){
                tbClient.updateProjectMember(Arrays.asList(UtilMap.getString(role, "userId")), tRoleIds, projectId);
            }
        }
        // 指定项目经理为编辑权限
        if (!pmUserId.isEmpty()) {
            tbClient.updateProjectMember(_convertUserId(pmUserId, false), Arrays.asList(pmRoleId), projectId);
        }
        // 同步项目分组 [prd 主数据建立产品与分组对照表，绑定末级分类接口，若未匹配则不添加]
        String groupName = UtilMap.getString(formData, "textField_m9z110mk");
        if (StringUtils.isNotBlank(groupName)) {
            List<Map> tagList = tbClient.queryTagList(UtilMap.map("q", groupName));
            String tagId;
            if (tagList.size() > 0) {
                tagId=UtilMap.getString(tagList.get(0), "id");
            }else{
                Map tag=tbClient.addTag(groupName,"");
                tagId=UtilMap.getString(tag,"id");
            }
            /// 更新项目分组, 不会保留原分组, 需要兼容
            List<String> tags = tbClient.queryProjectTag(projectId, tbConf.getOperatorId());
            tags.add(tagId);
            tbClient.updateProjectTag(projectId, tags, tbConf.getOperatorId());
        }
        // prd 人员变更, 自动同步更新执行人: 仅触发存在执行人 & 未完成场景 [5.10 新增项目不执行, TB任务号宜搭回写有延迟, 避免异常]
        if (!isCreate) {
            updateProjectRole(projectId, Arrays.asList("全部", "自动更新"), "", "");
        }
        // prd 5.23 更新项目概览信息 8.21 项目信息\项目重要性分级, 也添加差异同步, 避免新增字段历史项目不同步问题
//        if (UtilMap.isNotBlankString(formData, "src_desc")) {
//        if(pmUserId!=null&&pmUserId.size()>0){
//            tbClient.updateProjectStatusField(projectId, tbConf.getOperatorId(), TBConf.assembleCustomField("项目经理-IPD", _convertUserIdObj(pmUserId, false)));
//        }
//        if(zjUserId!=null&&zjUserId.size()>0){
//            tbClient.updateProjectStatusField(projectId, tbConf.getOperatorId(), TBConf.assembleCustomField("项目总监-IPD", _convertUserIdObj(zjUserId, false)));
//        }
//        }
//        if (UtilMap.isNotBlankString(formData, "src_level")) {
//            tbClient.updateProjectStatusField(projectId, tbConf.getOperatorId(), TBConf.assembleCustomFieldName("项目重要等级", UtilMap.getString(formData, "textField_lwj1r7n6")));
//        }
        // 项目添加到项目集
//        List<Map> groupList = tbClient.queryProgramList("",groupName);
//        Map item;
//        if (groupList!=null&&groupList.size()>0) {
//            item = groupList.get(0); // 权限
//        }else{
//            item = tbClient.createProgramList(groupName, "");
//        }
//        tbClient.upsertProgramProject(UtilMap.getString(item, "id"), Arrays.asList(projectId),tbConf.getOperatorId());
        /// 7.18 项目状态不一致执行概览字段更新, 避免tb重复更新造成操作记录
        if (UtilMap.isNotBlankString(formData, "src_state")) {
            tbClient.updateProjectStatusField(projectId, tbConf.getOperatorId(), TBConf.assembleCustomFieldName("状态", UtilMap.getString(formData, "textField_ltsdsti7")));
        }
        log.info("TB项目信息, {}, {}", projectId, roleIds.size());
    }

    // fixme 提取方法, 参考 getProjectCFID 实现
    private String _getProjectRoleId(String projectId, String roleName) {
        List<Map> roles = tbClient.queryProjectRoles(projectId);
        Optional optional = roles.stream().filter(item -> roleName.equals(item.get("name"))).findAny();
        McException.assertAccessException(!optional.isPresent(), roleName + ": 项目角色不存在");
        return UtilMap.getString((Map) optional.get(), "id");
    }

    @Override
    public void createTask(String formInstId) {
        // 查询项目主数据 [艾为应用]
        YDParam ydParam = YDParam.builder()
                .appType("APP_UUFCJROTRU87EZ0XWPV6")
                .systemToken("MT866991HJVPGXMV53O2OAVB0O2K39UVGBV2MTG1")
                .formUuid("FORM-ED6CACB45BB04C41BCD0962F54F7010FQQYG")
                .formInstanceId(formInstId)
                .build();
        // prd 客户系统主数据, 是全量定时从mdm更新, 因此不能增量
        DDR_New ddrNew=ydClient.queryData(ydParam, YDConf.FORM_QUERY.retrieve_id);
        Map formData = ddrNew.getFormData();
        log.info("项目主数据, {}", formData);
        McException.assertAccessException(formData.isEmpty(), "未匹配项目主数据");
        String projectCode=UtilMap.getString(formData,"textField_m8simbnk");
        List<Map> pList = ydService.queryDataList_FormData("FORM-87750D9F02094BCA88FAD8166F0B9AA7VJMG", UtilMap.map("textField_lrj7vnxb", projectCode)); // 先您过目编号匹配
        McException.assertAccessException(pList.isEmpty(), "未匹配到项目主数据");
        String projectId=UtilMap.getString(pList.get(0), "textField_lqxtykce"); // TB项目编号
        McException.assertAccessException(projectId==null||projectId.isEmpty(), "未匹配到TB项目编号");
        String taskType=_getTaskType(projectId,"风险");
        List<String> roleIds = (List<String>) UtilMap.getList(formData, "employeeField_m8fo7mpb_id");
        String taskId=saveTask(projectId,taskType,UtilMap.getString(formData,"textareaField_m9jmwi5l"),UtilMap.getString(formData,"textareaField_m9jmwi5q"),roleIds,"",null,null);
        String url="https://qekxvk.aliwork.com/APP_UUFCJROTRU87EZ0XWPV6/processDetail?formUuid=FORM-ED6CACB45BB04C41BCD0962F54F7010FQQYG&procInsId="+formInstId;
        _updateApproveField(taskId, new ArrayList<>(), url, "需求变更流程", "完成", AWServer.TASK_APPROVE_LINK);
        List<Map> table=UtilMap.getList(formData,"tableField_m8fo7mph");
        for (Map map:table){
            List<String> userIds = (List<String>) UtilMap.getList(map, "employeeField_m8zm84pm_id");
            saveTask(projectId,taskType,UtilMap.getString(map,"textField_m8fo7mpi"),UtilMap.getString(map,"textField_m8fo7mpj"),userIds,taskId,
                UtilMap.getLong(map,"dateField_m8fo7mpo"),UtilMap.getString(map,"textField_m8fo7mpl"));
        }
        ydClient.operateData(YDParam.builder()
                .appType("APP_UUFCJROTRU87EZ0XWPV6")
                .systemToken("MT866991HJVPGXMV53O2OAVB0O2K39UVGBV2MTG1")
                .formInstanceId(formInstId)
                .updateFormDataJson(JSON.toJSONString(UtilMap.map("textField_m9li9vzv", taskId)))
                .build(), YDConf.FORM_OPERATION.update);
        log.info("TB任务编号更新, {}, {}", formInstId, taskId);
    }

    @Override
    public void updateLcb(String projectCode) {
        YDParam ydParam = YDParam.builder()
                .appType("APP_CNQYZE26WA8UMSJJ188R")
                .systemToken("0O966HC1D4CQO63Q8IL1L5W6DV7D3UMXUHI3MPA")
                .formUuid("FORM-C34B18D925714348AED23A453A845D68VSPK")
                .build();
        if (StringUtils.isNotBlank(projectCode)) {
            ydParam.setSearchFieldJson(JSON.toJSONString(UtilMap.map("textField_m86rdjn0", projectCode)));
        }
        List<Map> mapBaseList = ydService.queryFormData_all(ydParam);
        if (StringUtils.isNotBlank(projectCode)) {
            mapBaseList = mapBaseList.stream().filter(item -> projectCode.equals(item.get("textField_m86rdjn0"))).collect(Collectors.toList());
        }
        log.info("项目里程碑数据, {}", mapBaseList.size());
        McException.assertAccessException(mapBaseList.isEmpty(), "未匹配到项目主数据");
        Map baseMap=mapBaseList.get(0);
        List<Map> pList = ydService.queryDataList_FormData("FORM-87750D9F02094BCA88FAD8166F0B9AA7VJMG", UtilMap.map("textField_lrj7vnxb", projectCode)); // 先您过目编号匹配
        McException.assertAccessException(pList.isEmpty(), "未匹配到项目主数据");
        String projectId=UtilMap.getString(pList.get(0), "textField_lqxtykce"); // TB项目编号
        // 搜索项目里程碑任务类型的id
        String taskTypeId=_getTaskType(projectId,"里程碑");
        // 搜索里程碑任务
        List<Map> taskList = tbClient.queryProjectTaskList(projectId, UtilMap.put(UtilMap.empty(), "q", "scenarioId = "+taskTypeId), null);
        // 修改截至时间
        for (Map task: taskList) {
            log.info("task:{}",task);
            String name=UtilMap.getString(task,"content");
            if(!LCB_YDFIELID.containsKey(name)) continue;
            String timeStm=UtilMap.getString(baseMap,LCB_YDFIELID.get(name));
            if(timeStm==null||timeStm.isEmpty()) continue;
            try {
                tbClient.updateTaskDueDate(UtilMap.getString(task,"id"),UtilDateTime.formatLocalDateTime(UtilDateTime.getLocalDateTimeFromTimestamp(Long.valueOf(timeStm))),tbConf.getOperatorId());
            }catch (Exception e){
                e.printStackTrace();
            }
        }
        tbClient.createGanttBaseline(projectId,UtilMap.getString(baseMap,"textField_m9s9wnfv"),null);
    }

    private static Map<String,String> LCB_YDFIELID;
    static {
        LCB_YDFIELID=new HashMap<>();
        LCB_YDFIELID.put("TR1","dateField_m86pxo3y");
        LCB_YDFIELID.put("CDCP","dateField_m86pxo45");
        LCB_YDFIELID.put("TR2","dateField_m86pxo3z");
        LCB_YDFIELID.put("TR3","dateField_m86pxo40");
        LCB_YDFIELID.put("PDCP","dateField_m86pxo44");
        LCB_YDFIELID.put("TR4","dateField_m86pxo4c");
        LCB_YDFIELID.put("TR4A","dateField_m86pxo4b");
        LCB_YDFIELID.put("TR5","dateField_m86pxo4a");
        LCB_YDFIELID.put("TR6","dateField_m86pxo49");
        LCB_YDFIELID.put("ADCP","dateField_m86pxo46");
    }

    private String saveTask(String projectId,String taskType,String content,String note,List<String> roleIds,String fId,Long dueDate,String effect){
        Map data=new HashMap();
        data.put("projectId",projectId);
        data.put("content",content);
        List<String> tbUserIds = _convertUserId(roleIds, false);
        // prd 多人情况避免, 手动指定后被覆盖. 判定当前任务执行人在角色范围内则不支持更新
        if (tbUserIds.size() > 0) {
            data.put("executorId",tbUserIds.get(0));// 执行人
        }
        data.put("note",note);// 备注
        data.put("scenariofieldconfigId",taskType);// 任务类型
        data.put("parentTaskId",fId);
        if(dueDate!=null){
            data.put("dueDate",UtilDateTime.formatLocalDateTime(UtilDateTime.getLocalDateTimeFromTimestamp(dueDate)));
        }
        Map taskMap=tbClient.createTask(tbConf.getOperatorId(),data);
        String taskId=UtilMap.getString(taskMap,"taskId");
        Map body = TBConf.assembleCustomFieldName("类型", "风险");
        tbClient.updateTaskCustomField(taskId, tbConf.getOperatorId(), body);
        if(effect!=null&&!effect.isEmpty()){
            Map body2 = TBConf.assembleCustomFieldName("预计实现效果", effect);
            tbClient.updateTaskCustomField(taskId, tbConf.getOperatorId(), body2);
        }
        return taskId;
    }

    /**
     * 分配项目角色 prd 若是一人直接指定, 多人情况下忽略
     */
    @Synchronized
    @Override
    public void updateProjectRole(String projectId, List<String> trNode, String production, String version) {

        // 项目档案 [tb项目号查询, 避免模糊匹配]
        List<Map> pList = ydService.queryDataList_FormData(ydIpdForm.getProject(), UtilMap.map("textField_lqxtykce", projectId));
        McException.assertAccessException(pList.isEmpty(), "未匹配到项目主数据");
        List<Map> rList = (List<Map>) pList.get(0).get("tableField_lqxtykcf");

        log.info("分配项目角色 projectId = {}, 项目数量 = {}, 角色列表 = {}", projectId, pList.size(), rList);
        // 任务编码字段ID
        List<Map> customField = tbClient.queryProjectCustomField(projectId, null);
        // prd 非未完成任务, 不执行人员匹配更新
        List<String> workFlowStatusList = _getWorkFlowStatusList(projectId, AWServer.WORKFLOW_INITIAL);

        //List<Map> taskList = tbClient.queryProjectTaskList(projectId, UtilMap.map("pageSize", 10), null);
        tbClient.queryProjectTaskList(projectId, null, taskList -> {
            for (Map task : taskList) {
                if (!workFlowStatusList.contains(task.get("tfsId"))) {
                    continue;
                }
//                // prd 0920 下发不阻断，没有执行人也更新，已下达 [ 存量任务兼容未添加字段 ]
//                try {
//                    Map body = TBConf.assembleCustomFieldName(AWServer.TASK_TRANSMIT, "已下达");
//                    tbClient.updateTaskCustomField(UtilMap.getString(task, "id"), tbConf.getOperatorId(), body);
//                } catch (McException e) {
//                    log.error(e.getMessage(), e);
//                }
                // 获取资源名称对应的项目角色
                Optional optional = customField.stream().filter(item -> AWServer.TASK_ROLE.equals(item.get("name"))).findAny();
                if (!optional.isPresent()) {
                    continue;
                }
                List<Map> customfields = (List<Map>) task.get("customfields");
                log.info("customfields:{}",customfields);
                log.info("customfields:{}",customField);
                if (!trNode.contains("全部")) {
                    Optional optional2 = customField.stream().filter(item -> AWServer.TASK_STAGE.equals(item.get("name"))).findAny();
                    if (!optional2.isPresent()) {
                        continue;
                    }
                    // prd 传递无TR评审节点值, 匹配空值, 作为分配条件
                    String trName = TBConf.getTaskFieldValue_First(customfields, UtilMap.getString((Map) optional2.get(), "id"));
                    if (StringUtils.isBlank(trName)) {
                        if (!trNode.contains(AWServer.TASK_STAGE_BLANK)) {
                            continue;
                        }
                    } else {
                        if (!trNode.contains(trName)) {
                            continue;
                        }
                    }
                    /// prd: 08.01 TR5 ~  TR6 有产品版本，为空则匹配所选 TR 下全部任务，否则按照所选产品、版本下达
                    if (UtilString.isNotBlankCompatNull(production)) {
                        Optional optional3 = customField.stream().filter(item -> AWServer.TASK_PRODUCT.equals(item.get("name"))).findAny();
                        if (!optional3.isPresent()) {
                            continue;
                        }
                    }
                    if (UtilString.isNotBlankCompatNull(version)) {
                        Optional optional3 = customField.stream().filter(item -> AWServer.TASK_PRODUCT_VERSION.equals(item.get("name"))).findAny();
                        if (!optional3.isPresent()) {
                            continue;
                        }
                    }
                    log.info("tr节点控制, {}, {}", trName, trNode);
                }
                // prd 人员变更, 自动同步更新执行人: 仅触发存在执行人 & 未完成场景
                if (trNode.contains("自动更新")) {
                    if (StringUtils.isBlank(UtilMap.getString(task, "executorId"))) {
                        continue;
                    }
                }
                String roleName = TBConf.getTaskFieldValue_First(customfields, UtilMap.getString((Map) optional.get(), "id"));
                // 获取角色在项目主数据对应成员
                optional = rList.stream().filter(item -> roleName.equals(item.get("selectField_lqxu6bgf"))).findAny();
                if (!optional.isPresent()) {
                    continue;
                }
                log.info("分配项目角色 taskId = {}, 资源名称 = {}, 角色信息 = {}", UtilMap.getString(task, "id"), roleName, optional.get());
                List<String> roleIds = (List<String>) UtilMap.getList((Map) optional.get(), "employeeField_lqxtykch_id");
                if (roleIds.size() == 1) {
                    String tbUserId = _convertUserId(roleIds.get(0), false);
                    if (StringUtils.isNotBlank(tbUserId)) {
                        tbClient.updateTaskExecutor(UtilMap.getString(task, "id"), tbConf.getOperatorId(), tbUserId, false, false);
                    }
                }
                // prd 6.17 任务分配：资源匹配多人时，取值第一人作为执行人，其他人员作为参与者
                if (roleIds.size() > 1) {
                    String executorId = UtilMap.getString(task, "executorId");
                    List<String> tbUserIds = _convertUserId(roleIds, false);
                    // prd 多人情况避免, 手动指定后被覆盖. 判定当前任务执行人在角色范围内则不支持更新
                    if (StringUtils.isNotBlank(executorId) && tbUserIds.contains(executorId)) {
                        log.info("多人情况跳过, {}", tbUserIds.contains(executorId));
                        continue;
                    }
                    if (tbUserIds.size() > 0) {
                        tbClient.updateTaskExecutor(UtilMap.getString(task, "id"), tbConf.getOperatorId(), tbUserIds.get(0), false, false);
                    }
                    tbClient.updateTaskInvolveMembers(UtilMap.getString(task, "id"), tbConf.getOperatorId(), tbUserIds, null, false, false);
                }
            }
            return true;
        });
    }

    // TB与宜搭userId转换 fixme 提取 tbService [ ppExt: 若TB未启用人员返回为空 ]
    private String _convertUserId(String userId, boolean isTBID) {
        List<Map> tbMap = tbClient.idMapQuery(userId, isTBID);
        // 过滤未匹配人员信息 [ppExt: TB人员未匹配, 不执行修改也无报错]
        if (tbMap.isEmpty()) {
            log.warn("TB与宜搭userId转换为空, userId = {}, tbID = {}", userId, isTBID);
            return "";
        }
        if (!isTBID) {
            return UtilMap.getString(tbMap.get(0), "tbId");
        } else {
            Map extra = UtilMap.getMap(tbMap.get(0), "extra");
            return UtilMap.getString(extra, "userId");
        }
    }

    // TB与宜搭userId转换 fixme 提取 tbService [ ppExt: 若TB未启用人员返回为空 ]
    private List<String> _convertUserId(List<String> userIds, boolean isTBID) {
        return userIds.stream().distinct()
                .map(id -> _convertUserId(id, isTBID))
                .filter(id -> StringUtils.isNotBlank(id))
                .collect(Collectors.toList());
    }

    private List<Object> _convertUserIdObj(List<String> userIds, boolean isTBID) {
        return userIds.stream().distinct()
                .map(id -> (Object) UtilMap.map("title, id","1",_convertUserId(id, isTBID)))
                .filter(obj -> StringUtils.isNotBlank(UtilMap.getString((Map) obj,"id")))
                .collect(Collectors.toList());
    }

    // 获取项目自定义字段ID, 格式 { 名称: id } fixme 提取
    Map<String, String> getProjectCFID(String projectId, List<String> names) {
        List<Map> customField = tbClient.queryProjectCustomField(projectId, null);
        List tList = customField.stream().filter(item -> names.contains(item.get("name"))).collect(Collectors.toList());
        return (Map<String, String>) tList.stream().reduce(UtilMap.empty(), (acc, cur) -> {
            ((Map) acc).put(((Map) cur).get("name"), ((Map) cur).get("id"));
            return acc;
        });
    }

    /// 修改任务自定义字段
    void _batchUpdate(String fieldName, String preName, String modifyName, String projectId) {
        // prd 非未完成任务字段ID
        List<String> workFlowStatusList = _getWorkFlowStatusList(projectId, AWServer.WORKFLOW_INITIAL);

        Map<String, String> proData = getProjectCFID(projectId, Arrays.asList(fieldName));
        String tql = "cf:" + proData.get(fieldName) + " = " + preName + " AND tfsId IN (" + String.join(", ", workFlowStatusList) + ")";

        List<Map> taskList = tbClient.queryProjectTaskList(projectId, UtilMap.put(UtilMap.empty(), "q", tql), null);
        for (Map task : taskList) {
            Map body = TBConf.assembleCustomFieldName(fieldName, modifyName);
            tbClient.updateTaskCustomField(UtilMap.getString(task, "id"), tbConf.getOperatorId(), body);
        }
    }

    /// fixme 权限成员列表，最大size30, 通过 partition 快速切割
    private void _batchUpdatePermission(List<String> userIds, String dentryUuid, String creatorUnionId, String roleId, boolean isDelete) {
        List<List<String>> parts = Lists.partition(userIds, DDConf.WORKSPACE_PERMISSION_SIZE);
        for (List<String> ids : parts) {
            List<Map> members = ids.stream().map(userId -> UtilMap.map("type, id", "USER", userId)).collect(Collectors.toList());
            if (isDelete) {
                ddClient_storage.deleteDentryPermissions(ddClient.getAccessToken(), dentryUuid, creatorUnionId, roleId, members, null);
            } else {
                ddClient_storage.updateDentryPermissions(ddClient.getAccessToken(), dentryUuid, creatorUnionId, roleId, members, null);
            }
        }
    }

    /// 更新一级归档目录 dentryUuid 权限打断, 并更新项目成员只读权限, 且更新记录实例ID [ 用于后续项目全量同步时, 刷新权限 ]
    private void _breakPermission(String rootNodeId, String dentryUuid, List<String> staffIds, String creatorUnionId, String formInstanceId) {

        ddClient_storage.updateDentryPermissionsInheritances(ddClient.getAccessToken(), dentryUuid, creatorUnionId, "BREAK");
        try {  // todo 目前权限绑定群聊会异常, BUG
            /// 原存量用户权限处理 & 管理者, 所有者更新异常处理 [ fixme: 需要获取知识库根目录 dentryUuid 成员列表作为权限原始数据 ]
            List<Map> allMembers = ddClient_storage.queryDentryPermissions(ddClient.getAccessToken(), rootNodeId, creatorUnionId, null);
            for (Map member : allMembers) {
                Map<String, String> team = UtilMap.getMap(member, "member");
                // ppExt: 管理员和所有者, 赋值编辑权限, 会报500异常需要兼容 [ 原只读权限不做调整 ]
                if ("USER".equals(UtilMap.getString(team, "type"))) {
                    String roleId = UtilMap.getString(UtilMap.getMap(member, "role"), "id");
                    String userId = UtilMap.getString(team, "id");
                    if (Arrays.asList("OWNER", "MANAGER", "READER").contains(roleId)) {
                        staffIds.remove(userId);
                    } else {
                        /// prd 存量的成员处理办法：第一次更新的时候刷为只读权限，后续知识库成员手动导致的变更不再单独添加。文件权限读取权限矩阵
                        if (StringUtils.isNotBlank(formInstanceId)) {
                            staffIds.add(userId); // ppExt: 原存量的用户, 改为查看权限 [兼容]
                        }
                    }
                }
            }
        } catch (McException e) {
            log.error(e.getMessage(), e);
        }
        // ppExt: 不移除 TBManager 编辑权限, 避免被覆盖
        if (staffIds.contains(ddConf.getOperator())) {
            staffIds.remove(ddConf.getOperator());
        }
        log.info("更新一级归档目录 {}, {}, {}", dentryUuid, staffIds, formInstanceId);
        _batchUpdatePermission(staffIds, dentryUuid, creatorUnionId, "READER", false);
        if (StringUtils.isNotBlank(formInstanceId)) {
            ydClient.operateData(YDParam.builder()
                    .formInstanceId(formInstanceId)
                    .updateFormDataJson(JSON.toJSONString(UtilMap.map("textField_lxmw6z7k", dentryUuid)))
                    .build(), YDConf.FORM_OPERATION.update);
        }
    }

    /// 同步知识库成员权限, 非初始化
    private void _assembleBreakPermission(Map formData) {
        List<String> staffIds = new ArrayList<>();
        String workspaceId = UtilMap.getString(formData, "textField_lx2yekqj");
        if (StringUtils.isNotBlank(workspaceId)) {
            /// ppExt 一级目录设置权限打断: 后续知识库成员调整, 非管理者与所有者人员不会传递到当前路径下, 即该文件若要可见需要重新对成员赋权
            String dentryUuid = UtilMap.getString(formData, "textField_lxmw6z7k");
            String creatorId = UtilMap.getString(formData, "textField_lx8ra1bs");
            List<Map> details = UtilMap.getList(formData, "tableField_lqxtykcf");
            for (Map detail : details) {
                if (ObjectUtil.isNotNull(detail.get("employeeField_lqxtykch_id"))) {
                    staffIds.addAll((List) detail.get("employeeField_lqxtykch_id"));
                }
            }
            String creatorUnionId = String.valueOf(ddClient_contacts.getUserInfoById(ddClient.getAccessToken(), creatorId).get("unionid"));
            _breakPermission(UtilMap.getString(formData, "textField_lx342bvd"), dentryUuid, staffIds, creatorUnionId, null);
        }
    }

    /// 获取产品主数据
    private List<Map> getProductData(String projectId, String q, String qKey, String tCode) {
        // 校验项目号
        YDSearch ytbProjectId = new YDSearch("textField_lqxtykce", projectId, "TB项目编号", YDSearch.Type.TEXT_FIELD, YDSearch.Operator.EQ);
        List<Map> tbProjectList = (List<Map>) ydClient.queryData(YDParam.builder().formUuid("FORM-141E21DF183846028E21727CE43CD1C75CLZ")
                .appType("APP_R5EBUF2FPN3Y8DRF93M4").systemToken("ON566NC1VNIHPANP9TNVHB3TBIWS3E0TUZ5RLF3")
                .searchCondition(JSONObject.toJSONString(Arrays.asList(ytbProjectId)))
                .pageSize(10).build(), YDConf.FORM_QUERY.retrieve_list).getData();

        McException.assertAccessException(tbProjectList == null || tbProjectList.size() < 1, "未匹配到主数据！");
        String pId = UtilMap.getString(UtilMap.getMap(tbProjectList.get(0), "formData"), "textField_lrj7vnxb");
        McException.assertAccessException(StringUtils.isBlank(pId), "未匹配到主数据！");

        YDSearch ydSearch = new YDSearch("textField_lha7mqbs", "G", "物料子类", YDSearch.Type.TEXT_FIELD, YDSearch.Operator.EQ);
        YDSearch ydSearch2 = new YDSearch("textField_lha7mqbn", pId, "项目号", YDSearch.Type.TEXT_FIELD, YDSearch.Operator.EQ);
        YDSearch ydSearch3 = new YDSearch(qKey, q, "产品料号", YDSearch.Type.TEXT_FIELD, YDSearch.Operator.LIKE);
        List qList = Arrays.asList(ydSearch, ydSearch2, ydSearch3);
        if (tCode != null) {
            YDSearch ydSearch4 = new YDSearch("textField_lhsxurnt", tCode, "产品料号", YDSearch.Type.TEXT_FIELD, YDSearch.Operator.EQ);
            qList = Arrays.asList(ydSearch, ydSearch2, ydSearch3, ydSearch4);
        }
        List<Map> list = (List<Map>) ydClient.queryData(YDParam.builder().formUuid("FORM-4W8667D17CAAGFXI9VO0J9J7RFNL2QDFL7AHLJ")
                .appType("APP_QBWQITQBSPJNYTUTNPDK").systemToken("8F966HB12J27MQJM6V4IQDYHYTPA2G4GTZGCLN1")
                .searchCondition(JSONObject.toJSONString(qList))
                .pageSize(100).build(), YDConf.FORM_QUERY.retrieve_list).getData();
        List<Map> result = new ArrayList<>();
        Map<String, String> keyMap = new HashMap<>();
        for (Map map : list) {
            String data = UtilMap.getString(UtilMap.getMap(map, "formData"), qKey);
            if (!StringUtils.isBlank(data) && !keyMap.containsKey(data)) {
                result.add(UtilMap.map("title", data));
                keyMap.put(data, data);
            }
        }
        return result;
    }

    @Override
    public void taskSatusUpdate(Map data) {
        String taskId = UtilMap.getString(data, "taskId");
        String projectId = UtilMap.getString(data, "projectId");
        YDParam ydParam = YDParam.builder()
                .appType("APP_UUFCJROTRU87EZ0XWPV6")
                .systemToken("MT866991HJVPGXMV53O2OAVB0O2K39UVGBV2MTG1")
                .formUuid("FORM-ED6CACB45BB04C41BCD0962F54F7010FQQYG")
                .searchCondition(JSONObject.toJSONString(Arrays.asList(
                        new YDSearch("textField_m9li9vzv", taskId, "TB任务编号", YDSearch.Type.TEXT_FIELD, YDSearch.Operator.EQ))))
                .build();
        List<Map> list =(List<Map>) ydClient.queryData(ydParam, YDConf.FORM_QUERY.retrieve_list).getData();
        log.info("查询到数据, {}", list.size());
        if(list==null||list.size()<1) return;
        String tfsName = _getWorkFlowStatusName(projectId, UtilMap.getString(data,"tfsId"));
        for (Map map:list){
            String no=UtilMap.getString(UtilMap.getMap(map,"formData"),"textField_m9gom6yl"); // 需求编号
            if(no == null || no.isEmpty()) continue;
            YDParam ydParam2 = YDParam.builder()
                    .appType("APP_UUFCJROTRU87EZ0XWPV6")
                    .systemToken("MT866991HJVPGXMV53O2OAVB0O2K39UVGBV2MTG1")
                    .formUuid("FORM-B3AB6C59BE434341AE0652F8F451F963WSDF")
                    .searchCondition(JSONObject.toJSONString(Arrays.asList(
                            new YDSearch("serialNumberField_m2u6t58x", no, "需求编号", YDSearch.Type.TEXT_FIELD, YDSearch.Operator.EQ))))
                    .build();
            // prd 客户系统主数据, 是全量定时从mdm更新, 因此不能增量
            List<Map> list2 =(List<Map>) ydClient.queryData(ydParam2, YDConf.FORM_QUERY.retrieve_list).getData();
            if(list2==null||list2.size()<1) return;
            String formInstId=UtilMap.getString(list2.get(0),"formInstanceId");
            ydClient.operateData(YDParam.builder()
                    .appType("APP_UUFCJROTRU87EZ0XWPV6")
                    .systemToken("MT866991HJVPGXMV53O2OAVB0O2K39UVGBV2MTG1")
                    .formInstanceId(formInstId)
                    .updateFormDataJson(JSON.toJSONString(UtilMap.map("textField_m9m8g98z, dateField_m9m8g990, textareaField_m9m8g991", tfsName
                    ,UtilDateTime.getLocalDateTimeTimeStamp(),"https://www.teambition.com/task/"+taskId)))
                            .useLatestVersion(true)
                    .build(), YDConf.FORM_OPERATION.update);
        }

    }

    @Override
    public Map doApprove(Map data) {
        boolean isChange=false;
        log.info("交付物审批, {}", data);
        String taskId = UtilMap.getString(data, "taskId");
        String title = UtilMap.getString(data, "title");

        // prd 618 通过任务扩展按钮触发, 接口仅需传递 taskId 即可, 相关信息直接从任务详情取值
        Map taskData = _getTaskFieldMap(taskId, AWServer.TASK_CODE,AWServer.TASK_APPROVE_ATTACHMENT);
        Map rTask = UtilMap.getMap(taskData, "task");
        String pCode = UtilMap.getString(rTask, "projectId");

        boolean isCallback = data.containsKey("tfsId");
        if (isCallback) {
            String projectId = UtilMap.getString(data, "projectId");
            String workFlowApprove = _getWorkFlowStatus(projectId, AWServer.WORKFLOW_APPROVE);
            if (!data.get("tfsId").equals(workFlowApprove)) {
                return null;
            }
        } else {
            // prd 发起评审先修改工作流，成功后再执行相关逻辑，避免前后置依赖与子任务未完成导致不能关闭
            String workFlowId_c = _getWorkFlowStatus(pCode, "已完成");
            if (!isChange) {
//                McException.assertAccessException(StringUtils.isBlank(UtilMap.getString(taskData, AWServer.TASK_CHECK_STATUS)), "请核对, 技术检查项检查状态!");
                // prd: 不过滤审批中, 避免撤销后重新填写交付件情况
                McException.assertAccessException(workFlowId_c.equals(rTask.get("tfsId")), "已完成任务, 不允许提交发布!");
                String workFlowId = _getWorkFlowStatus(pCode, "审批中");
                tbClient.updateTaskFlowStatus(taskId, tbConf.getOperatorId(), workFlowId, "审批中");
            } else {
                McException.assertAccessException(!rTask.get("tfsId").equals(workFlowId_c), "非已完成任务, 不允许提交变更!");
            }
        }
        List<Map> docs = _getDocs(taskData, AWServer.TASK_APPROVE_ATTACHMENT);
        // prd 7.10 交付件描述为空，即非交付件审批任务；且未传交付件内容，直接更新为审批通过。否则走交付物审批升版
//        if (UtilMap.isBlankString(taskData, AWServer.TASK_APPROVE_DESC) && docs.size() == 0) {

//        if (docs.size() == 0) {
//            String workFlowId = _getWorkFlowStatus(pCode, "审批通过-派智");
//            tbClient.updateTaskFlowStatus(taskId, tbConf.getOperatorId(), workFlowId, "审批通过-派智");
//            _updateApproveField(taskId, new ArrayList<>(), "", "非交付件任务", "已终止", AWServer.TASK_APPROVE_LINK);
//            return null;
//        }
        // ppExt: TB有卡控, 校验交付物不能为空, 同时兼容非交付件任务类型
        McException.assertAccessException(!isCallback && docs.size() == 0, "材料填写不能为空, 请填写后再操作!");
        String executorId = UtilMap.getString(rTask, "executorId");
        if (!isCallback) {
            McException.assertAccessException(StringUtils.isBlank(executorId), "任务执行人未分配!");
        }

        // 返回userId, 若非执行人权限, 页面控制弹出框
        String userId = "";
        String tCode = UtilMap.getString(taskData, AWServer.TASK_CODE);
        // 工作流tb虽然配置了流转逻辑, 冗余稳定性
        List<Map> formList = (List<Map>) ydClient.queryData(YDParam.builder()
                .formUuid(ydIpdForm.getReview())
                .searchFieldJson(JSON.toJSONString(UtilMap.map("textField_m9w625cs, textField_lrncs2fu", tCode, pCode)))
                .instanceStatus("RUNNING")
                .build(), YDConf.FORM_QUERY.retrieve_search_process).getData();
        if (formList.size() > 0 && formList.stream().filter(item -> tCode.equals(((Map) item.get("data")).get("textField_m9w625cs"))).findAny().isPresent()) {
            McException.assertAccessException(true, "已存在审批中流程, 请勿重复提交!");
        }
        String result = "";
        List<Map> tList = null;
        Map rProject = null;
        if (StringUtils.isBlank(tCode)) {
            result = "无任务编码";
        } else {
            List<Map> pList = ydService.queryDataList_FormData(ydIpdForm.getProject(), UtilMap.map("textField_lqxtykce", pCode));
            pList = pList.stream().filter(item -> pCode.equals(item.get("textField_lqxtykce"))).collect(Collectors.toList());
            if (pList.size() == 0) {
                result = "未匹配到项目主数据";
            } else {
                rProject = pList.get(0);
                // prd 多模板适配: [任务号 + 项目类型]
                String proType = UtilMap.getString(rProject, "textField_ltwcq7s6");
                tList = ydService.queryDataList_FormData(ydIpdForm.getApprove(), UtilMap.map("textField_lrnbx77u, radioField_lrnddfq6, selectField_m9zkr3qv", tCode, "启用",proType));
                tList = tList.stream().filter(item -> tCode.equals(item.get("textField_lrnbx77u")) ).collect(Collectors.toList()); // && UtilMap.getList(item, "multiSelectField_ltwjre9s").contains(proType)
                if (tList.size() == 0) {
                    result = "未配置交付物审批";
                }
            }
            log.info("交付物结果, {}", result);
            /// 发起评审 prd 630 若无交付物审批, 则配置发起人自动通过, 以兼容全部任务类型可提交交付件评审
            boolean isIgnore = "未配置交付物审批".equals(result);
            if (StringUtils.isBlank(result) || isIgnore) {
                result = ""; // 重置返回错误信息
                List<Map> roles = (List<Map>) rProject.get("tableField_lqxtykcf");
                Map formData = UtilMap.map("textField_m9w625cs, textField_lrncs2fu", tCode, pCode);
                formData.put("attachmentField_lqxtebtq", docs);
                formData.put("textField_m9zo9yci", UtilMap.getString(rTask, "content"));                            // 任务标题
                formData.put("selectField_lqxuswze", UtilMap.getString(rProject, "textField_lrj7vnxb"));            // 项目编号
                formData.put("textField_lqxuc9m4", UtilMap.getString(rProject, "textField_m9sfk19j"));          // 项目名称
                formData.put("employeeField_m9zo9ycj", UtilMap.getString(rProject, "employeeField_ltzn872j_id"));   // 项目经理 0402 控制矩阵角色为空, 流转到PM
                formData.put("textField_ltwcq7s6", UtilMap.getString(rProject, "textField_ltwcq7s6"));              // 项目类型
//                formData.put("textareaField_lw0nmvko", UtilMap.getString(taskData, AWServer.TASK_APPROVE_DESC));         // 交付件描述
//                formData.put("textField_lvbrueqs", UtilMap.getString(taskData, AWServer.TASK_PRODUCT));                  // 产品型号
//                formData.put("textField_lvbrueqt", UtilMap.getString(taskData, AWServer.TASK_PRODUCT_VERSION));          // 产品版本
                formData.put("radioField_ly2o9rxn", isIgnore ? "是" : "否");                        // 审批分支
                // 匹配任务编码与项目角色
                List<Map<String, String>> compIds = Arrays.asList(  // 任务表角色, 交付物评审表: 角色, 审批人
                        UtilMap.map("tsRole, prRole, prEmp", "multiSelectField_lrokzlo7, multiSelectField_lrokzlo7, employeeField_lqxtebtw"),
                        UtilMap.map("tsRole, prRole, prEmp", "multiSelectField_lrokzlo8, multiSelectField_lrokzlo8, employeeField_lqxtebtx"),
                        UtilMap.map("tsRole, prRole, prEmp", "multiSelectField_lrokzlo9, multiSelectField_lrokzlo9, employeeField_lqxtebty"));
                if (!isIgnore) {
                    Map rCondition = tList.get(0);
                    for (Map compId : compIds) {
                        // ppExt: 宜搭多选若无值则不返回字段 [prd 角色多选, 会签逻辑]
                        if (!rCondition.containsKey(compId.get("tsRole"))) {
                            continue;
                        }
                        List<String> namesCondition = (List<String>) rCondition.get(compId.get("tsRole"));
                        List<String> namesApproveRole = new ArrayList<>();
                        List<String> namesApprover = new ArrayList<>();
                        for (String name : namesCondition) {
                            Optional optional = roles.stream().filter(item -> UtilString.equal(name, item.get("selectField_lqxu6bgf"))).findAny();
                            if (optional.isPresent()) {
                                Map detail = (Map) optional.get();
                                if (ObjectUtil.isNotNull(detail.get("selectField_lqxu6bgf"))) {
                                    namesApproveRole.add(String.valueOf(detail.get("selectField_lqxu6bgf")));
                                }
                                if (ObjectUtil.isNotNull(detail.get("employeeField_lqxtykch_id"))) {
                                    namesApprover.addAll((List) detail.get("employeeField_lqxtykch_id"));
                                }
                            }
                        }
                        formData.put(compId.get("prRole"), namesApproveRole);
                        formData.put(compId.get("prEmp"), namesApprover);
                    }
                }
                // 组装数据
                Map<String, String> extra = (Map) tbClient.idMapQuery(executorId, "dingTalk-user", ddConf.getCorpId()).get(0).get("extra");
                UtilMap.putAll(formData, UtilMap.map("textField_lr3dlwsa, textField_lr3er4qb", taskId, executorId));
                formData.put("employeeField_lui5fu7z", Arrays.asList(extra.get("userId")));  // 0402 指定发起人, 用于结构化数据子流程发起人
//                formData.put("radioField_lxkaihe5", type);
//                formData.put("selectField_lxl7xtwe", isChange ? "提交变更" : "提交发布");       // 评审类型标题
                userId = extra.get("userId");
                // prd 618 通过任务扩展按钮触发, 变更填写关联pcr后再提交, 返回数据. 并需要回调
                if (!isChange) {
                    String instanceId = (String) ydClient.operateData(YDParam.builder()
                            .formUuid(ydIpdForm.getReview())
                            .processCode(ydIpdForm.getReviewProcess())
                            .formDataJson(JSON.toJSONString(formData))
                            .userId(extra.get("userId"))
                            .build(), YDConf.FORM_OPERATION.start);
                    result = ydIpdForm.getDomain() + ydConf.getAppType() + "/processDetail?procInsId=" + instanceId;
                } else {
                    formData.put("result", result);
                    formData.put("userId", userId);
                    List<String> pmUserIds = (List<String>) JSON.parse(UtilMap.getString(rProject, "employeeField_ltzn872j_id"));
                    if (ObjectUtil.isNotNull(pmUserIds)) {
                        formData.put("PMName", ddClient_contacts.getUserInfoById(ddClient.getAccessToken(), pmUserIds.get(0)).get("name"));
                    }
                    return formData;
                }
            }
        }
        // prd 618 通过任务扩展按钮, 提交后再触发审批回写, 直接提交场景下直接更新审批链接
        if (!isChange) {
            _updateApproveField(taskId, new ArrayList<>(), result, title, "审批中", AWServer.TASK_APPROVE_LINK);
        }
        return UtilMap.map("result, userId", result, userId);
    }

    @Autowired
    private YDConf ydConf;

    @Override
    public void checked(Map data) {
        log.info("检查项回调, {}", data);
        String title = UtilMap.getString(data, "title");
        Map body = TBConf.assembleCustomFieldName(AWServer.TASK_CHECK_STATUS, "已检查");
        tbClient.updateTaskCustomField(String.valueOf(data.get("taskId")), tbConf.getOperatorId(), body);
        // prd 7.31 复制的任务若是检查项已提交的任务, 也要触发 7.17 预检项任务ID与当前任务ID不匹配 逻辑
        String result = ydIpdForm.getDomain() + ydConf.getAppType() + "/formDetail?formInstId=" + data.get("formInstId") + "&taskId=" + UtilMap.getString(data, "taskId");
        Map body2 = TBConf.assembleCustomFieldName(AWServer.TASK_CHECK_LINK, result);
        tbClient.updateTaskCustomField(String.valueOf(data.get("taskId")), tbConf.getOperatorId(), body2);
//        _updateApproveField(UtilMap.getString(data, "taskId"), new ArrayList<>(), result, title, "已提交", AWServer.TASK_CHECK_LINK);
    }

    /**
     * 交付物审批回调
     */
    @Override
    public void approved(Map data) {

        log.info("交付物审批回调, {}", data);
//        String type = UtilMap.getString(data, "type");
        String projectId = UtilMap.getString(data, "projectId");
        String taskId = UtilMap.getString(data, "taskId");
        String title = UtilMap.getString(data, "title");
        String instanceId = UtilMap.getString(data, "instanceId");
        String result = UtilMap.getString(data, "approve").contains("通过") ? "审批通过" : "审批拒绝";
//        if ("变更".equals(type)) {
//            Map body = TBConf.assembleCustomFieldName(AWServer.TASK_APPROVE_STATE, "已变更");
//            tbClient.updateTaskCustomField(taskId, tbConf.getOperatorId(), body);
//        }
//        // prd 618 兼容原逻辑, 保留回调类更新任务工作流. 新版本通过任务扩展按钮触发, 发布在创建时触发\变更不更新工作流避免导致逾期情况
//        if ("回调".equals(type)) {
//            String workFlowId = _getWorkFlowStatus(projectId, result);
//            tbClient.updateTaskFlowStatus(taskId, tbConf.getOperatorId(), workFlowId, result);
//        }
//        if ("发布".equals(type)) {
            String workFlowId = _getWorkFlowStatus(projectId, result);
            tbClient.updateTaskFlowStatus(taskId, tbConf.getOperatorId(), workFlowId, result);
//        }
        String url = ydIpdForm.getDomain() + ydConf.getAppType() + "/processDetail?procInsId=" + instanceId;
        _updateApproveField(taskId, new ArrayList<>(), url, title, result, AWServer.TASK_APPROVE_LINK);
    }

    /// 重置任务状态
    void resetWorkFlow(String projectId, String taskId) {
        List<String> workIds = _getWorkFlowStatusList(projectId, AWServer.WORKFLOW_INITIAL);
        for (String workFlowId : workIds) {
            boolean isSuccess = false;
            try {
                tbClient.updateTaskFlowStatus(taskId, tbConf.getOperatorId(), workFlowId, "一二级任务不允许N/A");
                isSuccess = true;
            } catch (McException e) {

            }
            if (isSuccess) {
                break;
            }
        }
    }
}



